#include <fstream>
#include <cstdio>
#include "api.h"
#include "wingui.h"
#include "nls.h"
#include "osdep.h"
#include "enumerator.h"
#include "game.h"
#include "loaderstate.h"
#include "action.h"
#include "race.h"
#include "program.h"
#include "util.h"

class GameDirProvider : public ValueEnumerator<Game> {
    Data& d;
    HWND wnd;
 public:
    GameDirProvider(Data& d, HWND wnd)
        : d(d), wnd(wnd)
        { }
    void handleItem(Game& g) {
        if (g.getLocalDirectory().empty()) {
            std::string n = d.getNewGameDirectory(g.getName(), false);
            if (n.empty()) {
                int result = MessageBox(wnd,
                                        (_("I cannot allocate a Winplan slot for ") + g.getName()
                                        + _(".\r\n"
                                            "You can let me assign a directory outside the Winplan tree, which is "
                                            "then only accessible to non-Winplan programs. "
                                            "Or, leave this directory unassigned, make room by deleting other "
                                            "games, and invoke the 'assign' function later to assign this game "
                                            "its Winplan slot.\r\n"
                                            "Do you want me to assign a directory outside the Winplan tree?")).c_str(),
                                        _("Game Directory"),
                                        MB_ICONSTOP | MB_YESNO);
                if (result == IDNO)
                    return;
                n = d.getNewGameDirectory(g.getName(), true);
            }
            g.setLocalDirectory(n);
        }
    }
};

MainWindow::MainWindow(Data& data)
    : m_Hwnd(0),
      m_Data(data),
      m_FileMenu(0), m_GameMenu(0), m_RaceMenu(0), m_ActionMenu(0), m_ProgramMenu(0),
      m_HelpMenu(0), m_FullMenu(0),
      m_Terminated(false)
{
    m_BigFont = CreateFont(24, 0, 0, 0, FW_BOLD, TRUE /*italic*/, FALSE, FALSE, DEFAULT_CHARSET,
                           OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS, DEFAULT_QUALITY, FF_SWISS,
                           "Arial");
    m_SmallFont = CreateFont(16, 0, 0, 0, FW_BOLD, FALSE /*italic*/, FALSE, FALSE, DEFAULT_CHARSET,
                             OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS, DEFAULT_QUALITY, FF_SWISS,
                             "Arial");
}

static void
drawText(HDC dc, int x, int& y, std::string text)
{
    TEXTMETRIC tm;
    GetTextMetrics(dc, &tm);

    RECT r;
    r.left = x;
    r.right = x + 10000; // FIXME
    r.top = y;
    r.bottom = y + tm.tmHeight;
    DrawText(dc, text.data(), text.size(), &r, DT_LEFT);

    y = r.bottom + 1;
}

LRESULT
MainWindow::handle(HWND hwnd, UINT msg, WPARAM wparam, LPARAM lparam)
{
    switch (msg) {
     case WM_CLOSE:
        m_Terminated = true;
        return 0;
     case WM_COMMAND:
        switch (UINT cmd = LOWORD(wparam)) {
         case ID_FILE_CLOSE:
            m_Terminated = true;
            return 0;
         case ID_FILE_LOAD:
            loadFile();
            return 0;
         case ID_FILE_ASSIGN_DIRS:
            {
                GameDirProvider gdp(m_Data, hwnd);
                m_Data.enumerateGames(gdp);
                InvalidateRect(hwnd, 0, TRUE);
                rebuildAllMenus();
            }
            return 0;
         case ID_HELP_ABOUT:
            {
                std::string s = _("VGAP Integrator");
                s += " ";
                s += VERSION;
                s += "\r\n\r\n";
                s += _("Compiled ");
                s += __DATE__;
                s += " ";
                s += __TIME__;
                MessageBox(hwnd, s.c_str(), _("About"), MB_OK);
            }
            return 0;
         default:
            if (cmd >= ID_GAME && cmd < ID_GAME + m_Games.size()) {
                selectGame(m_Games[cmd - ID_GAME]);
                if (m_Selection.getSelectedGame()->getLocalDirectory().size() == 0) {
                    MessageBox(hwnd, _("This game does not yet have a local directory assigned. "
                                       "Most functions do not work without one. "
                                       "Use File->Assign... to assign it a directory."),
                               _("Game Directory"),
                               MB_ICONSTOP | MB_OK);
                }
            } else if (cmd >= ID_RACE && cmd < ID_RACE + m_Races.size()) {
                selectRace(m_Races[cmd - ID_RACE]);
            } else if (cmd >= ID_ACTION && cmd < ID_ACTION + m_Actions.size()) {
                runAction(m_Actions[cmd - ID_ACTION]);
            } else if (cmd >= ID_PROGRAM && cmd < ID_PROGRAM + m_ProgramActions.size()) {
                runAction(m_ProgramActions[cmd - ID_PROGRAM]);
            }
            return 0;
        }
     case WM_PAINT: {
        PAINTSTRUCT ps;
        HDC dc = BeginPaint(hwnd, &ps);
        HGDIOBJ oldfont = SelectObject(dc, m_BigFont);
        int y = 30;
        int x = 30;
        if (Game* g = m_Selection.getSelectedGame())
            drawText(dc, x, y, g->getName());
        else
            drawText(dc, x, y, _("No game selected."));

        y += 10;
        SelectObject(dc, m_SmallFont);
        if (Race* r = m_Selection.getSelectedRace()) {
            drawText(dc, x, y, _("Player ") + itoa(r->getRaceNumber()));
        } else {
            drawText(dc, x, y, _("No player selected."));
        }
        if (Game* g = m_Selection.getSelectedGame()) {
            if (g->getLocalDirectory().size())
                drawText(dc, x, y, _("Game data in ") + g->getLocalDirectory());
            else
                drawText(dc, x, y, _("No game directory assigned."));
        }
        SelectObject(dc, oldfont);
        EndPaint(hwnd, &ps);
        return 0;
     }
     default:
        return Window::handle(hwnd, msg, wparam, lparam);
    }
}

void
MainWindow::clearMenu(HMENU& menu)
{
    if (menu == 0) {
        menu = CreateMenu();
    } else {
        // TODO: prfen...
        do {
            HMENU m = GetSubMenu(menu, 0);
            if (m != 0)
                DestroyMenu(m);
        } while (RemoveMenu(menu, 0, MF_BYPOSITION));
    }
}

void
MainWindow::buildFileMenu()
{
    clearMenu(m_FileMenu);
    AppendMenu(m_FileMenu, MF_STRING, ID_FILE_LOAD, _("&Load..."));
    AppendMenu(m_FileMenu, MF_STRING, ID_FILE_ASSIGN_DIRS, _("&Assign missing game directories"));
    AppendMenu(m_FileMenu, MF_STRING, ID_FILE_CLOSE, _("&Exit"));
}

void
MainWindow::buildGameMenu()
{
    clearMenu(m_GameMenu);

    m_Games.clear();
    VectorValueEnumerator<Game> e(m_Games);
    m_Data.enumerateGames(e);

    // HACK
    if (!m_Selection.getSelectedGame() && m_Games.size())
        m_Selection.selectGame(m_Games.front());

    for (unsigned int i = 0; i < m_Games.size(); ++i) {
        int flags;
        if (m_Selection.getSelectedGame() == m_Games[i])
            flags = MF_STRING | MF_CHECKED;
        else
            flags = MF_STRING;
        AppendMenu(m_GameMenu, flags, ID_GAME + i, m_Games[i]->getName().c_str());
    }

    if (GetMenuItemCount(m_GameMenu) == 0)
        AppendMenu(m_GameMenu, MF_STRING | MF_GRAYED, ID_GAME, "(empty)");
}

void
MainWindow::buildRaceMenu()
{
    clearMenu(m_RaceMenu);

    m_Races.clear();
    VectorValueEnumerator<Race> e(m_Races);
    if (Game* g = m_Selection.getSelectedGame())
        g->enumerateRaces(e);

    for (unsigned int i = 0; i < m_Races.size(); ++i) {
        int flags;
        if (m_Selection.getSelectedRace() == m_Races[i])
            flags = MF_STRING | MF_CHECKED;
        else
            flags = MF_STRING;

        char tmp[200];
        std::sprintf(tmp, _("Race %d"), m_Races[i]->getRaceNumber());

        AppendMenu(m_RaceMenu, flags, ID_RACE + i, tmp);
    }

    if (GetMenuItemCount(m_RaceMenu) == 0)
        AppendMenu(m_RaceMenu, MF_STRING | MF_GRAYED, ID_RACE, "(empty)");
}

void
MainWindow::buildActionMenu()
{
    clearMenu(m_ActionMenu);

    m_Actions.clear();
    VectorValueEnumerator<Action> a(m_Actions);
    // FIXME:
    if (Game* g = m_Selection.getSelectedGame())
        m_Data.enumerateSimpleActions(*g, a);

    for (unsigned int i = 0; i < m_Actions.size(); ++i) {
        int flags;
        m_Selection.selectAction(m_Actions[i]);
        if (m_Selection.expandString(m_Actions[i]->getCommandLine()).size())
            flags = MF_STRING;
        else
            flags = MF_STRING | MF_GRAYED;

        AppendMenu(m_ActionMenu, flags, ID_ACTION + i,
                   (m_Actions[i]->getDescription() + _(" (using ") + m_Actions[i]->getProgram().getName() + ")").c_str());
    }

    if (GetMenuItemCount(m_ActionMenu) == 0)
        AppendMenu(m_ActionMenu, MF_STRING | MF_GRAYED, ID_ACTION, "(empty)");
}

void
MainWindow::buildProgramsMenu()
{
    clearMenu(m_ProgramMenu);

    std::vector<Program*> progs;
    VectorValueEnumerator<Program> a(progs);
    m_Data.enumeratePrograms(a);

    m_ProgramActions.clear();
    VectorValueEnumerator<Action> aa(m_ProgramActions);

    for (unsigned int i = 0; i < progs.size(); ++i) {
        // enumerate this program's actions, and store in array
        unsigned int org = m_ProgramActions.size();
        progs[i]->enumerateActions(aa);
        unsigned int count = m_ProgramActions.size() - org;

        HMENU the_menu;
        std::string pfx, sfx;
        if (count <= 1) {
            the_menu = m_ProgramMenu;
            pfx = progs[i]->getName() + " (";
            sfx = ")";
        } else {
            the_menu = CreateMenu();
            AppendMenu(m_ProgramMenu, MF_POPUP, UINT(the_menu), progs[i]->getName().c_str());
        }

        while (org < m_ProgramActions.size()) {
            int flags;
            m_Selection.selectAction(m_ProgramActions[org]);
            if (m_Selection.expandString(m_ProgramActions[org]->getCommandLine()).size())
                flags = MF_STRING;
            else
                flags = MF_STRING | MF_GRAYED;

            AppendMenu(the_menu, flags, ID_PROGRAM + org,
                       (pfx + m_ProgramActions[org]->getDescription() + sfx).c_str());
            ++org;
        }
    }

    if (GetMenuItemCount(m_ProgramMenu) == 0)
        AppendMenu(m_ProgramMenu, MF_STRING | MF_GRAYED, ID_PROGRAM, "(empty)");
}

void
MainWindow::buildHelpMenu()
{
    clearMenu(m_HelpMenu);
    AppendMenu(m_HelpMenu, MF_STRING, ID_HELP_ABOUT, "&About...");
}

void
MainWindow::rebuildAllMenus()
{
    buildGameMenu();
    buildRaceMenu();
    buildActionMenu();
    buildProgramsMenu();
}

void
MainWindow::selectGame(Game* g)
{
    m_Selection.selectGame(g);
    buildGameMenu();
    buildRaceMenu();
    buildActionMenu();
    buildProgramsMenu();
    InvalidateRect(m_Hwnd, 0, TRUE);
}

void
MainWindow::selectRace(Race* r)
{
    m_Selection.selectRace(r);
    buildRaceMenu();
    buildActionMenu();
    buildProgramsMenu();
    InvalidateRect(m_Hwnd, 0, TRUE);
}

void
MainWindow::runAction(Action* a)
{
    m_Selection.selectAction(a);

    std::string command = m_Selection.expandString(a->getCommandLine());
    MessageBox(m_Hwnd, ("This would run the command line\r\n" + command).c_str(),
               a->getDescription().c_str(), MB_OK);
}

bool
MainWindow::askYesNo(const char* format, int arg)
{
    char tmp[300];
    std::sprintf(tmp, format, arg);
    return MessageBox(m_Hwnd, tmp, "Load Definition File", MB_YESNO | MB_ICONQUESTION) == IDYES;
}

void
MainWindow::loadFile(std::string name)
{
    std::ifstream in(name.c_str());
    if (!in) {
        MessageBox(m_Hwnd, (_("Unable to open file '") + name + "'").c_str(),
                   _("Load Definition File"), MB_ICONSTOP | MB_OK);
    } else {
        try {
            Data tmp;
            LoaderState st;
            st.setBaseDirectory(getDirectoryName(name));
            tmp.read(in, st);

            if (st.getNumItems(st.Programs) == 0 && st.getNumItems(st.Games) == 0) {
                MessageBox(m_Hwnd, _("File does not contain any usable information."),
                           _("Load Definition File"), MB_ICONSTOP | MB_OK);
            } else {
                st.setAccept(st.Globals,  false);
                st.setAccept(st.Programs, st.getNumItems(st.Programs) != 0 &&
                             askYesNo(_("Accept %d program definitions?"), st.getNumItems(st.Programs)));
                st.setAccept(st.Games,    st.getNumItems(st.Games)    != 0 &&
                             askYesNo(_("Accept %d game definitions?"), st.getNumItems(st.Games)));
                if (st.isAcceptAny()) {
                    in.clear(); // weil wir am Ende waren
                    in.seekg(0);
                    m_Data.read(in, st);

                    if (st.isAccept(st.Games)) {
                        GameDirProvider gdp(m_Data, m_Hwnd);
                        m_Data.enumerateGames(gdp);
                    }

                    rebuildAllMenus();
                }
            }
        }
        catch (std::exception& e) {
            MessageBox(m_Hwnd,
                       (std::string(_("Load aborted due to error:\r\n")) + e.what()).c_str(),
                       _("Load Definition File"),
                       MB_ICONSTOP + MB_OK);
        }
    }
}

void
MainWindow::loadFile()
{
    char buffer[300];
    OPENFILENAME ofn;
    memset(&ofn, 0, sizeof(ofn));
    buffer[0] = 0;
    ofn.lStructSize          = sizeof(ofn);
    ofn.hwndOwner            = m_Hwnd;
    ofn.lpstrFilter          = "VGA Planets Game Definition Files (*.vgd)\0*.vgd\0"
                               "VGA Planets Program Definition Files (*.vpd)\0*.vpd\0"
                               "All Files\0*.*\0";
    // ofn.lpstrCustomFilter = 0;
    // ofn.nMaxCustFilter    = 0;
    // ofn.nFilterIndex      = 0;
    ofn.lpstrFile            = buffer;
    ofn.nMaxFile             = sizeof(buffer);
    // ofn.lpstrFileTitle    = 0;
    // ofn.nMaxFileTitle     = 0;
    // ofn.lpstrInitialDir   = 0;
    ofn.lpstrTitle           = _("Load Definition File");
    ofn.Flags                = OFN_FILEMUSTEXIST;
    // ofn.nFileOffset       = 0;
    // ofn.nFileExtension    = 0;
    // ofn.lpstrDefExt       = 0;

    if (GetOpenFileName(&ofn))
        loadFile(buffer);
}


HMENU
MainWindow::getMenu()
{
    if (m_FullMenu == 0) {
        buildFileMenu();
        buildGameMenu();
        buildRaceMenu();
        buildActionMenu();
        buildProgramsMenu();
        buildHelpMenu();

        m_FullMenu = CreateMenu();
        AppendMenu(m_FullMenu, MF_POPUP, UINT(m_FileMenu), _("&File"));
        AppendMenu(m_FullMenu, MF_POPUP, UINT(m_GameMenu), _("&Game"));
        AppendMenu(m_FullMenu, MF_POPUP, UINT(m_RaceMenu), _("&Race"));
        AppendMenu(m_FullMenu, MF_POPUP, UINT(m_ActionMenu), _("&Action"));
        AppendMenu(m_FullMenu, MF_POPUP, UINT(m_ProgramMenu), _("&Programs"));
        AppendMenu(m_FullMenu, MF_POPUP, UINT(m_HelpMenu), _("&Help"));
    }
    return m_FullMenu;
}

static void
loadDefaults(Data& d)
{
    std::string name = getSettingsFileName();
    std::ifstream in(name.c_str());
    if (in) {
        LoaderState st;
        st.setBaseDirectory(getDirectoryName(name));
        d.read(in, st);
    }
}

static void
saveDefaults(Data& d)
{
    std::string filename = getSettingsFileName();
    std::string bak = filename + ".bak";
    std::remove(bak.c_str());
    std::rename(filename.c_str(), bak.c_str());
    std::ofstream out(filename.c_str());
    if (out)
        d.write(out);
}

int
RealMain(int argc, char** argv)
{
    Window::registerWindowClass("MyWindowClass", 0);
    Data data;
    loadDefaults(data);
    MainWindow mainwin(data);
    HWND wnd = mainwin.create("MyWindowClass", "VGAP Integrator", WS_OVERLAPPEDWINDOW,
                        CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT,
                        0, mainwin.getMenu());
    mainwin.setWindowHandle(wnd);
    ShowWindow(wnd, SW_SHOW);

    MSG msg;
    while (!mainwin.isTerminated() && GetMessage(&msg, 0, 0, 0)) {
        TranslateMessage(&msg);
        DispatchMessage(&msg);
    }

    saveDefaults(data);

    return 0;
}
